from typing import List
from dataclasses import dataclass, field
from tensorflow_caney.model.config.cnn_config import Config


@dataclass
class CloudMaskConfig(Config):

    # model parameters from Huggingface
    hf_repo_id: str = 'nasa-cisto-data-science-group/vhr-cloudmask'
    hf_model_filename: str = 'cloudmask-vietnam-senegal-46-0.04.hdf5'

    # test classes definition for validation
    test_classes: List[str] = field(
        default_factory=lambda: ['no-cloud', 'cloud', 'thin-cloud']
    )

    # rewrite default output to be COG
    prediction_driver: str = 'COG'

    # postprocessing settings
    postprocessing_steps: List[str] = field(
        default_factory=lambda: ['sieve', 'smooth', 'fill', 'dilate']
    )



from vhr_cloudmask.model.pipelines.cloudmask_cnn_pipeline \
    import CloudMaskPipeline

__author__ = """Jordan A. Caraballo-Vega"""
__email__ = "jordan.a.caraballo-vega@nasa.gov"
__version__ = "1.3.2"

__all__ = ["CloudMaskPipeline"]



import os
import re
import cv2
import time
import logging
import rasterio
import numpy as np
import xarray as xr
import rioxarray as rxr
import rasterio.features as riofeat

from pathlib import Path
from scipy import ndimage
from skimage import color
from omegaconf import OmegaConf
from huggingface_hub import hf_hub_download

from tensorflow_caney.utils.data import modify_bands, \
    get_mean_std_metadata, read_metadata
from tensorflow_caney.utils import indices
from tensorflow_caney.inference import inference
from tensorflow_caney.utils.model import load_model
from tensorflow_caney.utils.system import seed_everything

from vhr_cloudmask.model.config.cloudmask_config \
    import CloudMaskConfig as Config
from tensorflow_caney.model.pipelines.cnn_segmentation import CNNSegmentation

CHUNKS = {'band': 'auto', 'x': 'auto', 'y': 'auto'}
__status__ = "Production"


# -----------------------------------------------------------------------------
# class CloudMaskPipeline
# -----------------------------------------------------------------------------
class CloudMaskPipeline(CNNSegmentation):
    """This is a conceptual class representation of a CNN Segmentation
    TensorFlow pipeline. It is essentially an extended combination of the
    :class:`tensorflow_caney.model.pipelines.cnn_segmentation.CNNSegmentation`.

    :param logger: A logger device
    :type logger: str
    :param conf: Configuration device
    :type conf: omegaconf.OmegeConf object
    :param data_csv: CSV filename with data files for training
    :type data_csv: str
    :param experiment_name: Experiment name description
    :type experiment_name: str
    :param images_dir: Directory to store training images
    :type images_dir: str
    :param labels_dir: Directory to store training labels
    :type labels_dir: str
    :param model_dir: Directory to store trained models
    :type model_dir: str
    """
    # -------------------------------------------------------------------------
    # __init__
    # -------------------------------------------------------------------------
    def __init__(
                self,
                config_filename: str = None,
                data_csv: str = None,
                model_filename: str = None,
                output_dir: str = None,
                inference_regex_list: str = None,
                input_bands: str = None,
                output_bands: str = None,
                postprocessing_steps: list = None,
                default_config: str = 'templates/cloudmask_default.yaml',
                logger=None
            ):
        """Constructor method
        """
        logging.info('Initializing CloudMaskPipeline')

        # Configuration file intialization
        if config_filename is None:
            config_filename = os.path.join(
                os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
                default_config)
            logging.info(f'Loading default config: {config_filename}')

        # load configuration into object
        self.conf = self._read_config(config_filename, Config)

        # Set logger
        self.logger = logger if logger is not None else self._set_logger()

        # rewrite model filename option if given from CLI
        if model_filename is not None:
            assert os.path.exists(model_filename), \
                f'{model_filename} does not exist.'
            self.conf.model_filename = model_filename

        # rewrite output directory if given from CLI
        if output_dir is not None:
            self.conf.inference_save_dir = output_dir
            os.makedirs(self.conf.inference_save_dir, exist_ok=True)

        # rewrite inference regex list
        if inference_regex_list is not None:
            self.conf.inference_regex_list = inference_regex_list

        # rewrite input bands if given from CLI
        if input_bands is not None:
            self.conf.input_bands = input_bands

        # rewrite output bands if given from CLI
        if output_bands is not None:
            self.conf.output_bands = output_bands

        # rewrite postprocessing steps if given from CLI
        if postprocessing_steps is not None:
            self.conf.postprocessing_steps = postprocessing_steps

        # Set Data CSV
        self.data_csv = data_csv

        # Set experiment name
        try:
            self.experiment_name = self.conf.experiment_name.name
        except AttributeError:
            self.experiment_name = self.conf.experiment_name

        # Output directories for metadata
        self.images_dir = os.path.join(self.conf.data_dir, 'images')
        logging.info(f'Images dir: {self.images_dir}')

        self.labels_dir = os.path.join(self.conf.data_dir, 'labels')
        logging.info(f'Labels dir: {self.labels_dir}')

        self.model_dir = self.conf.model_dir
        logging.info(f'Model dir:  {self.labels_dir}')

        logging.info(f'Output dir: {self.conf.inference_save_dir}')

        # Create output directories
        for out_dir in [
                self.images_dir, self.labels_dir,
                self.model_dir]:
            os.makedirs(out_dir, exist_ok=True)

        # save configuration into the model directory
        try:
            OmegaConf.save(
                self.conf, os.path.join(self.model_dir, 'config.yaml'))
            logging.info(f'Saved config under: {self.model_dir}')
        except PermissionError:
            logging.info('No permissions to save config, skipping step.')

        logging.info(
            f'Output GeoTIFF driver: {self.conf.prediction_driver}')

        # Seed everything
        seed_everything(self.conf.seed)

    # -------------------------------------------------------------------------
    # predict
    # -------------------------------------------------------------------------
    def predict(self) -> None:
        """This will perform inference on a list of GeoTIFF files provided
        as a list of regexes from the CLI.

        :return: None, outputs GeoTIFF cloudmask files to disk.
        :rtype: None
        """

        logging.info('Starting prediction stage')

        # if model filename does not exist, load the default model from HF
        if not os.path.exists(self.conf.model_filename):
            logging.info(
                f'{self.conf.model_filename} does not exist. ' +
                'Dowloading default model from HuggingFace.'
            )
            model_filename = hf_hub_download(
                repo_id=self.conf.hf_repo_id,
                filename=self.conf.hf_model_filename)
        else:
            model_filename = self.conf.model_filename
        logging.info(f'Model filename: {model_filename}')

        # Load model for inference
        model = load_model(
            model_filename=model_filename,
            model_dir=self.model_dir,
            conf=self.conf
        )

        # Retrieve mean and std, there should be a more ideal place
        if self.conf.standardization in ["global", "mixed"]:
            mean, std = get_mean_std_metadata(
                os.path.join(
                    self.model_dir,
                    f'mean-std-{self.conf.experiment_name}.csv'
                )
            )
            logging.info(f'Mean: {mean}, Std: {std}')
        else:
            mean = None
            std = None

        # gather metadata
        if self.conf.metadata_regex is not None:
            metadata = read_metadata(
                self.conf.metadata_regex,
                self.conf.input_bands,
                self.conf.output_bands
            )

        # Gather filenames to predict
        if len(self.conf.inference_regex_list) > 0:
            data_filenames = self.get_filenames(self.conf.inference_regex_list)
        else:
            data_filenames = self.get_filenames(self.conf.inference_regex)
        logging.info(f'{len(data_filenames)} files to predict')

        # iterate files, create lock file to avoid predicting the same file
        for filename in sorted(data_filenames):

            # start timer
            start_time = time.time()

            # set output directory
            basename = os.path.basename(os.path.dirname(filename))
            if basename == 'M1BS' or basename == 'P1BS':
                basename = os.path.basename(
                    os.path.dirname(os.path.dirname(filename)))

            output_directory = os.path.join(
                self.conf.inference_save_dir, basename)
            os.makedirs(output_directory, exist_ok=True)

            # set prediction output filename
            output_filename = os.path.join(
                output_directory,
                f'{Path(filename).stem}.{self.conf.experiment_type}.tif')

            # lock file for multi-node, multi-processing
            lock_filename = f'{output_filename}.lock'

            # predict only if file does not exist and no lock file
            if not os.path.isfile(output_filename) and \
                    not os.path.isfile(lock_filename):

                try:

                    logging.info(f'Starting to predict {filename}')

                    # if metadata is available
                    if self.conf.metadata_regex is not None:

                        # get timestamp from filename
                        year_match = re.search(
                            r'(\d{4})(\d{2})(\d{2})', filename)
                        timestamp = str(int(year_match.group(2)))

                        # get monthly values
                        mean = metadata[timestamp]['median'].to_numpy()
                        std = metadata[timestamp]['std'].to_numpy()
                        self.conf.standardization = 'global'

                    # create lock file
                    open(lock_filename, 'w').close()

                    # open filename
                    image = rxr.open_rasterio(filename)

                    # handle NTF files here, that differ from GeoTIFFs
                    if Path(filename).suffix == '.ntf':
                        image = image[0].to_array().squeeze()

                    logging.info(f'Prediction shape: {image.shape}')

                    # check bands in imagery, do not proceed if one band
                    if image.shape[0] == 1:
                        logging.info(
                            'Skipping file because of non sufficient bands')
                        continue

                except rasterio.errors.RasterioIOError:
                    logging.info(f'Skipped {filename}, probably corrupted.')
                    continue

                # Calculate indices and append to the original raster
                image = indices.add_indices(
                    xraster=image, input_bands=self.conf.input_bands,
                    output_bands=self.conf.output_bands)

                # Modify the bands to match inference details
                image = modify_bands(
                    xraster=image, input_bands=self.conf.input_bands,
                    output_bands=self.conf.output_bands)
                logging.info(f'Prediction shape after modf: {image.shape}')

                logging.info(
                    f'Prediction min={image.min().values}, ' +
                    f'max={image.max().values}')

                # Transpose the image for channel last format
                image = image.transpose("y", "x", "band")

                # Remove no-data values to account for edge effects
                temporary_tif = xr.where(image > -100, image, 600)

                # Sliding window prediction
                prediction, probability = \
                    inference.sliding_window_tiler_multiclass(
                        xraster=temporary_tif,
                        model=model,
                        n_classes=self.conf.n_classes,
                        overlap=self.conf.inference_overlap,
                        batch_size=self.conf.pred_batch_size,
                        threshold=self.conf.inference_treshold,
                        standardization=self.conf.standardization,
                        mean=mean,
                        std=std,
                        normalize=self.conf.normalize,
                        rescale=self.conf.rescale,
                        window=self.conf.window_algorithm,
                        probability_map=self.conf.probability_map
                    )

                # apply default postprocessing
                prediction = self.postprocessing(
                    prediction, self.conf.postprocessing_steps)

                # get cloud metrics for metadata
                cloud_metadata = self.calc_metadata(prediction)

                # Drop image band to allow for a merge of mask
                image = image.drop(
                    dim="band",
                    labels=image.coords["band"].values[1:],
                )

                # Get metadata to save raster prediction
                prediction = xr.DataArray(
                    np.expand_dims(prediction, axis=-1),
                    name=self.conf.experiment_type,
                    coords=image.coords,
                    dims=image.dims,
                    attrs=image.attrs
                )

                # Add metadata to raster attributes
                prediction.attrs['long_name'] = (self.conf.experiment_type)
                prediction.attrs['model_name'] = (model_filename)

                # add cloud metadata to raster attributes
                for mkey, mvalue in cloud_metadata.items():
                    prediction.attrs[mkey] = (mvalue)

                # transpose prediction for saving output
                prediction = prediction.transpose("band", "y", "x")

                # Set nodata values on mask
                nodata = prediction.rio.nodata
                prediction = prediction.where(image != nodata)
                prediction.rio.write_nodata(
                    self.conf.prediction_nodata, encoded=True, inplace=True)

                # Save output raster file to disk
                prediction.rio.to_raster(
                    output_filename,
                    BIGTIFF="IF_SAFER",
                    compress=self.conf.prediction_compress,
                    driver=self.conf.prediction_driver,
                    dtype=self.conf.prediction_dtype
                )
                del prediction

                # save probability map
                if probability is not None:

                    probability = xr.DataArray(
                        np.expand_dims(probability, axis=-1),
                        name=self.conf.experiment_type,
                        coords=image.coords,
                        dims=image.dims,
                        attrs=image.attrs
                    )

                    # Add metadata to raster attributes
                    probability.attrs['long_name'] = (
                        self.conf.experiment_type)
                    probability.attrs['model_name'] = (
                        model_filename)
                    probability = probability.transpose("band", "y", "x")

                    # Set nodata values on mask
                    nodata = probability.rio.nodata
                    probability = probability.where(image != nodata)
                    probability.rio.write_nodata(
                        self.conf.prediction_nodata,
                        encoded=True, inplace=True
                    )

                    # Save output raster file to disk
                    probability.rio.to_raster(
                        Path(output_filename).with_suffix('.prob.tif'),
                        BIGTIFF="IF_SAFER",
                        compress=self.conf.prediction_compress,
                        driver=self.conf.prediction_driver,
                        dtype='float32'
                    )
                    del probability

                # delete lock file
                try:
                    os.remove(lock_filename)
                except FileNotFoundError:
                    logging.info(f'Lock file not found {lock_filename}')
                    continue

                logging.info(f'Finished processing {output_filename}')
                logging.info(f"{(time.time() - start_time)/60} min")

            # This is the case where the prediction was already saved
            else:
                logging.info(f'{output_filename} is done or lock file exists.')

        return

    # -------------------------------------------------------------------------
    # postprocessing
    # -------------------------------------------------------------------------
    def postprocessing(
                self,
                prediction: np.ndarray,
                postprocessing_steps: list = []
            ) -> np.ndarray:

        # sieve clearing of small objects
        if 'sieve' in postprocessing_steps:
            riofeat.sieve(prediction, 800, prediction, None, 8)

        # cloud smoothing
        if 'smooth' in postprocessing_steps:
            prediction = ndimage.median_filter(prediction, size=20)

        # binary fill holes
        if 'fill' in postprocessing_steps:
            prediction = ndimage.binary_fill_holes(prediction).astype(int)

        # dilate
        if 'dilate' in postprocessing_steps:

            # get prediction and set cv2 format
            prediction = np.uint8(np.squeeze(prediction) * 255)
            _, thresh = cv2.threshold(prediction, 127, 255, 0)

            # gather contours
            contours, _ = cv2.findContours(
                thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)

            # set output contours
            conts = np.zeros((thresh.shape[0], thresh.shape[1], 3))

            # iterate over each contour and increase size
            for c in contours:
                c = self.scale_contour(c, 1.12)  # 1.05 looks decent
                conts = cv2.drawContours(
                    conts, [c], -1, (0, 255, 0), cv2.FILLED)

            # converting mask into raster
            prediction = color.rgb2gray(conts).astype('int16')
            prediction[prediction > 0] = 1

        return prediction

    # -------------------------------------------------------------------------
    # calc_metadata
    # -------------------------------------------------------------------------
    def calc_metadata(self, prediction: np.ndarray) -> dict:

        # set metadata directory for cloud metadata
        metadata_dict = {}

        #  cloud percentage
        metadata_dict['pct_cloudcover_total'] = self.calc_cloud_percentage(
            prediction)

        return metadata_dict

    # -------------------------------------------------------------------------
    # calc_metadata
    # -------------------------------------------------------------------------
    def calc_cloud_percentage(self, prediction: np.ndarray) -> float:

        # get unique values per class
        unique, counts = np.unique(prediction, return_counts=True)
        unique_dict = dict(zip(unique, counts))

        # get cloudy pixels
        try:
            non_cloud_pixels = unique_dict[0]
        except (IndexError, KeyError):
            non_cloud_pixels = 0

        # get cloudy pixels
        try:
            cloud_pixels = unique_dict[1]
        except (IndexError, KeyError):
            cloud_pixels = 0

        # percent cloud cover
        pct_cloud_cover = round(
            100 * (cloud_pixels / (non_cloud_pixels + cloud_pixels)), 2)

        return pct_cloud_cover

    # -------------------------------------------------------------------------
    # scale_contour
    # -------------------------------------------------------------------------
    def scale_contour(self, cnt: np.ndarray, scale: float) -> np.ndarray:
        M = cv2.moments(cnt)
        try:
            cx = int(M['m10']/M['m00'])
            cy = int(M['m01']/M['m00'])
        except ZeroDivisionError:
            cx, cy = 0, 0

        cnt_norm = cnt - [cx, cy]
        cnt_scaled = cnt_norm * scale
        cnt_scaled = cnt_scaled + [cx, cy]
        return cnt_scaled.astype(np.int32)



# -*- coding: utf-8 -*-
"""
CPU and GPU Random Forest Pipeline - preprocess, train, predict and visualize.
Author: Jordan A. Caraballo-Vega, jordan.a.caraballo-vega@nasa.gov
"""

import os
import gc
import sys
import glob
import random
import logging
import argparse
from typing import List
from pathlib import Path
from tqdm import tqdm

import cv2
import joblib
import numpy as np
import xarray as xr
import pandas as pd
import rioxarray as rxr
from scipy.ndimage import median_filter, binary_fill_holes

from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier as sklRFC

from sklearn.metrics import accuracy_score, \
    precision_score, recall_score, f1_score

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

# from model.indices import addindices, fdi, si, ndwi, modify_bands

try:
    import cupy as cp
    import cudf as cf
    from cuml.ensemble import RandomForestClassifier as cumlRFC
    from cuml.dask.ensemble import RandomForestClassifier as cumlRFC_mg
    from cupyx.scipy.ndimage import median_filter_gpu
    cp.random.seed(seed=None)
    HAS_GPU = True
except ImportError:
    HAS_GPU = False

__author__ = "Jordan A Caraballo-Vega, Science Data Processing Branch"
__email__ = "jordan.a.caraballo-vega@nasa.gov"
__status__ = "Production"

__all__ = ["cumlRFC_mg", "median_filter_gpu", "median_filter"]


# -----------------------------------------------------------------------------
# rf_driver.py methods
# -----------------------------------------------------------------------------
def modify_bands(
        xraster: xr.core.dataarray.DataArray, input_bands: List[str],
        output_bands: List[str], drop_bands: List[str] = []):
    """
    Drop multiple bands to existing rasterio object
    """
    # Do not modify if image has the same number of output bands
    if xraster['band'].shape[0] == len(output_bands):
        return xraster

    # Drop any bands from input that should not be on output
    for ind_id in list(set(input_bands) - set(output_bands)):
        drop_bands.append(input_bands.index(ind_id)+1)
    return xraster.drop(dim="band", labels=drop_bands, drop=True)


def predict(data, model, ws=[5120, 5120], probabilities=False):
    """
    Predict from model.
    :param data: raster xarray object
    :param model: loaded model object
    :param ws: window size to predict on
    :return: prediction output in numpy format
    ----------
    Example
        raster.toraster(filename, raster_obj.prediction, outname)
    ----------
    """
    # open rasters and get both data and coordinates
    rast_shape = data[0, :, :].shape  # shape of the wider scene
    wsy, wsx = ws[0], ws[1]  # in memory sliding window predictions

    # if the window size is bigger than the image, predict full image
    if wsy > rast_shape[0]:
        wsy = rast_shape[0]
    if wsx > rast_shape[1]:
        wsx = rast_shape[1]

    prediction = np.zeros(rast_shape)  # crop out the window
    logging.info(f'wsize: {wsy}x{wsx}. Prediction shape: {prediction.shape}')

    for sy in tqdm(range(0, rast_shape[0], wsy)):  # iterate over x-axis
        for sx in range(0, rast_shape[1], wsx):  # iterate over y-axis
            y0, y1, x0, x1 = sy, sy + wsy, sx, sx + wsx  # assign window
            if y1 > rast_shape[0]:  # if selected x exceeds boundary
                y1 = rast_shape[0]  # assign boundary to x-window
            if x1 > rast_shape[1]:  # if selected y exceeds boundary
                x1 = rast_shape[1]  # assign boundary to y-window

            window = data[:, y0:y1, x0:x1]  # get window
            window = window.stack(z=('y', 'x'))  # stack y and x axis
            window = window.transpose("z", "band").values  # reshape

            # perform sliding window prediction
            if probabilities:
                prediction[y0:y1, x0:x1] = \
                    model.predict_proba(window)[:, 1].reshape(
                        (y1 - y0, x1 - x0))
            else:
                prediction[y0:y1, x0:x1] = \
                    model.predict(window).reshape((y1 - y0, x1 - x0))
    # save raster
    return prediction


def grow(merged_mask, eps=120):
    struct = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (eps, eps))
    return cv2.morphologyEx(merged_mask, cv2.MORPH_CLOSE, struct)


def denoise(merged_mask, eps=30):
    struct = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (eps, eps))
    return cv2.morphologyEx(merged_mask, cv2.MORPH_OPEN, struct)


def binary_fill(merged_mask):
    return binary_fill_holes(merged_mask)


# -----------------------------------------------------------------------------
# main
#
# python rf_driver.py options here
# -----------------------------------------------------------------------------
def main():

    # Process command-line args.
    desc = 'Random Forest Segmentation pipeline for tabular and spatial data.'
    parser = argparse.ArgumentParser(description=desc)

    parser.add_argument(
        '--gpu', dest='has_gpu', action='store_true', default=False)

    parser.add_argument(
        '--output-model', type=str, default='output.pkl', required=False,
        dest='output_pkl', help='Path to the output PKL file (.pkl)')

    parser.add_argument(
        '--data-csv', type=str, required=False,
        dest='data_csv', help='Path to the data CSV configuration file')

    parser.add_argument(
        '--train-csv', type=str, required=False,
        dest='train_csv', help='Path to the output CSV file')

    parser.add_argument(
        '--bands', type=str, nargs='*', required=False,
        dest='bands', help='Bands to store in CSV file',
        default=['CoastalBlue', 'Blue', 'Green', 'Yellow',
                 'Red', 'RedEdge', 'NIR1', 'NIR2'])

    parser.add_argument(
        '--input-bands', type=str, nargs='*', required=False,
        dest='input_bands', help='Input bands from predicted data',
        default=['CoastalBlue', 'Blue', 'Green', 'Yellow',
                 'Red', 'RedEdge', 'NIR1', 'NIR2'])

    parser.add_argument(
        '--step', type=str, nargs='*', required=True,
        dest='pipeline_step', help='Pipeline step to perform',
        default=['preprocess', 'train', 'predict', 'vis'],
        choices=['preprocess', 'train', 'predict', 'vis'])

    parser.add_argument(
        '--seed', type=int, required=False, dest='seed',
        default=42, help='Random SEED value')

    parser.add_argument(
        '--test-size', type=float, required=False,
        dest='test_size', default=0.20, help='Test size rate (e.g .30)')

    parser.add_argument(
        '--n-trees', type=int, required=False,
        dest='n_trees', default=20, help='Number of trees (e.g 20)')

    parser.add_argument(
        '--max-features', type=str, required=False,
        dest='max_feat', default='log2', help='Max features (e.g log2)')

    parser.add_argument(
        '--rasters', type=str, required=False, dest='rasters',
        default='*.tif', help='rasters to search for')

    parser.add_argument(
        '--window-size', type=int, required=False,
        dest='ws', default=5120, help='Prediction window size (e.g 5120)')

    parser.add_argument(
        '--output-dir', type=str, required=False,
        dest='output_dir', default='', help='output directory')

    parser.add_argument(
        '--output-probabilities', dest='output_probabilities',
        action='store_true', default=False)

    parser.add_argument(
        '--postprocess', type=bool, required=False,
        dest='postprocess', default=True,
        help='bool for postprocessing (denoise, fill)')

    args = parser.parse_args()

    # --------------------------------------------------------------------------------
    # set logging
    # --------------------------------------------------------------------------------
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    ch = logging.StreamHandler(sys.stdout)  # set stdout handler
    ch.setLevel(logging.INFO)

    # set formatter and handlers
    formatter = logging.Formatter(
        "%(asctime)s; %(levelname)s; %(message)s", "%Y-%m-%d %H:%M:%S")
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    # --------------------------------------------------------------------------------
    # preprocess step
    # --------------------------------------------------------------------------------
    if "preprocess" in args.pipeline_step:

        # ----------------------------------------------------------------------------
        # 1. Read data csv file
        # ----------------------------------------------------------------------------
        assert os.path.exists(args.data_csv), f'{args.data_csv} not found.'
        data_df = pd.read_csv(args.data_csv)  # initialize data dataframe
        assert not data_df.isnull().values.any(), f'NaN found: {args.data_csv}'
        logging.info(f'Open {args.data_csv} for preprocessing.')

        # ----------------------------------------------------------------------------
        # 2. Extract points out of spatial imagery - rasters
        # ----------------------------------------------------------------------------
        points_df = pd.DataFrame(columns=args.bands + ['CLASS'])
        points_list = []
        logging.info(f"Generating {data_df['ntiles'].sum()} points dataset.")

        # start iterating over each file
        for df_index in data_df.index:

            # get filename for output purposes, in the future, add to column
            filename = Path(data_df['data'][df_index]).stem
            logging.info(f'Processing {filename}')

            # read imagery from disk and process both image and mask
            img = rxr.open_rasterio(data_df['data'][df_index]).values
            mask = rxr.open_rasterio(data_df['label'][df_index]).values
            logging.info(f'Imagery shape: {img.shape}')

            # squeeze mask if needed, all non-1 values to 0
            mask = np.squeeze(mask) if len(mask.shape) != 2 else mask

            # get N points from imagery
            points_per_class = data_df['ntiles'][df_index] // 2

            # extract values from imagery, two classes
            for cv in range(2):

                bbox = img.shape  # size of the image
                counter = 0  # counter for class balancing

                while counter < points_per_class:

                    # get indices and extract spectral and class value
                    y, x = random.randrange(bbox[1]), random.randrange(bbox[2])
                    sv, lv = img[:, y, x], int(mask[y, x])

                    if lv == cv:
                        # trying speed up here - looks like from list is faster
                        points_list.append(
                            pd.DataFrame(
                                [np.append(sv, [lv])],
                                columns=list(points_df.columns))
                        )
                        counter += 1

        df_points = pd.concat(points_list)

        # ----------------------------------------------------------------------------
        # 3. Save file to disk
        # ----------------------------------------------------------------------------
        os.makedirs(Path(args.train_csv).parent.absolute(), exist_ok=True)
        df_points.to_csv(args.train_csv, index=False)
        logging.info(f'Saved dataset file at: {args.train_csv}')

    # --------------------------------------------------------------------------------
    # train step
    # --------------------------------------------------------------------------------
    if "train" in args.pipeline_step:

        # ----------------------------------------------------------------------------
        # 1. Read data csv file
        # ----------------------------------------------------------------------------
        assert os.path.exists(args.train_csv), f'{args.train_csv} not found.'
        data_df = pd.read_csv(
            args.train_csv, sep=',', names=args.bands + ['CLASS'])
        assert not data_df.isnull().values.any(), f'Na found: {args.train_csv}'
        logging.info(f'Open {args.train_csv} dataset for training.')

        # ----------------------------------------------------------------------------
        # 2. Shuffle and Split Dataset
        # ----------------------------------------------------------------------------
        data_df = data_df.sample(frac=1).reset_index(drop=True)  # shuffle data

        # split dataset, fix type
        x = data_df.iloc[:, :-1].astype(np.float32)
        y = data_df.iloc[:, -1].astype(np.int8)

        # split data into training and test
        x_train, x_test, y_train, y_test = train_test_split(
            x, y, test_size=args.test_size, random_state=args.seed)
        del data_df, x, y

        # logging some of the model information
        logging.info(f'X size: {x_train.shape}, Y size:  {y_train.shape}')
        logging.info(f'ntrees={str(args.n_trees)}, maxfeat={args.max_feat}')

        # ------------------------------------------------------------------
        # 3. Instantiate RandomForest object - FIX this area
        # ------------------------------------------------------------------
        if args.has_gpu:  # run using RAPIDS library

            # initialize cudf data and log into GPU memory
            logging.info('Training model via RAPIDS.')

            # single gpu setup
            x_train = cf.DataFrame.from_pandas(x_train)
            x_test = cf.DataFrame.from_pandas(x_test)
            y_train = cf.Series(y_train.values)
            rf_funct = cumlRFC  # RF Classifier

        else:
            logging.info('Training model via SKLearn.')
            rf_funct = sklRFC

        # Initialize model
        rf_model = rf_funct(
            n_estimators=args.n_trees, max_features=args.max_feat)

        # ------------------------------------------------------------------
        # 4. Fit Model
        # ------------------------------------------------------------------
        # fit model to training data and predict for accuracy score
        rf_model.fit(x_train, y_train)

        if args.has_gpu:
            acc_score = accuracy_score(
                y_test, rf_model.predict(x_test).to_array())
            p_score = precision_score(
                y_test, rf_model.predict(x_test).to_array())
            r_score = recall_score(
                y_test, rf_model.predict(x_test).to_array())
            f_score = f1_score(
                y_test, rf_model.predict(x_test).to_array())
        else:
            acc_score = accuracy_score(y_test, rf_model.predict(x_test))
            p_score = precision_score(y_test, rf_model.predict(x_test))
            r_score = recall_score(y_test, rf_model.predict(x_test))
            f_score = f1_score(y_test, rf_model.predict(x_test))

        logging.info(f'Test Accuracy:  {acc_score}')
        logging.info(f'Test Precision: {p_score}')
        logging.info(f'Test Recall:    {r_score}')
        logging.info(f'Test F-Score:   {f_score}')

        # make output directory
        os.makedirs(
            os.path.dirname(os.path.realpath(args.output_pkl)), exist_ok=True)

        # export model to file
        try:
            joblib.dump(rf_model, args.output_pkl)
            logging.info(f'Model has been saved as {args.output_pkl}')
        except Exception as e:
            logging.error(f'ERROR: {e}')

    # --------------------------------------------------------------------------------
    # predict step
    # --------------------------------------------------------------------------------
    if "predict" in args.pipeline_step:

        assert os.path.exists(args.output_pkl), f'{args.output_pkl} not found.'
        model = joblib.load(args.output_pkl)  # loading pkl in parallel
        logging.info(f'Loaded model {args.output_pkl}.')

        # create output directory
        os.makedirs(args.output_dir, exist_ok=True)

        # 3b3. apply model and get predictions
        rasters = glob.glob(args.rasters)
        assert len(rasters) > 0, "No raster found"
        logging.info(f'Predicting {len(rasters)} files.')

        for raster_filename in rasters:  # iterate over each raster

            filename = Path(raster_filename).stem
            output_filename = os.path.join(
                args.output_dir, f'{filename}.cloudmask.tif')

            if not os.path.isfile(output_filename):

                gc.collect()  # clean garbage
                logging.info(f'Starting new prediction...{raster_filename}')
                img = rxr.open_rasterio(raster_filename)

                img = modify_bands(
                    img, input_bands=args.input_bands,
                    output_bands=args.bands)
                logging.info(f'Imagery shape after band mod {img.shape}')

                prediction = predict(
                    img, model, ws=[args.ws, args.ws],
                    probabilities=args.output_probabilities)
                print(prediction.min(), prediction.max())

                if args.postprocess and not args.output_probabilities:
                    prediction = denoise(np.uint8(prediction))
                    prediction = binary_fill(prediction)
                    prediction = grow(np.uint8(prediction))

                img = img.transpose("y", "x", "band")
                img = img.drop(
                    dim="band",
                    labels=img.coords["band"].values[1:],
                    drop=True
                )

                prediction = xr.DataArray(
                    np.expand_dims(prediction, axis=-1),
                    name='mask',
                    coords=img.coords,
                    dims=img.dims,
                    attrs=img.attrs)

                prediction.attrs['long_name'] = ('mask')
                prediction = prediction.transpose("band", "y", "x")

                nodata = prediction.rio.nodata
                prediction = prediction.where(img != nodata)
                prediction.rio.write_nodata(nodata, encoded=True, inplace=True)
                prediction.rio.to_raster(
                    output_filename, BIGTIFF="IF_SAFER", compress='LZW')

                del prediction

            else:
                logging.info(f'{output_filename} already predicted.')

    # --------------------------------------------------------------------------------
    # predict step
    # --------------------------------------------------------------------------------
    if "vis" in args.pipeline_step:

        assert os.path.exists(args.output_pkl), f'{args.output_pkl} not found.'
        model = joblib.load(args.output_pkl)  # loading pkl in parallel
        logging.info(f'Loaded model {args.output_pkl}.')

    return


# -----------------------------------------------------------------------------
# Invoke the main
# -----------------------------------------------------------------------------
if __name__ == "__main__":
    sys.exit(main())



import sys
import time
import logging
import argparse
from vhr_cloudmask.model.pipelines.cloudmask_cnn_pipeline import \
    CloudMaskPipeline


# -----------------------------------------------------------------------------
# main
#
# python cloudmask_pipeline_cli.py -c config.yaml -d config.csv -s preprocess
# -----------------------------------------------------------------------------
def main():

    # Process command-line args.
    desc = 'Use this application to perform CNN segmentation.'
    parser = argparse.ArgumentParser(description=desc)

    parser.add_argument('-c',
                        '--config-file',
                        type=str,
                        required=False,
                        default=None,
                        dest='config_file',
                        help='Path to the configuration file')

    parser.add_argument('-d',
                        '--data-csv',
                        type=str,
                        required=False,
                        default=None,
                        dest='data_csv',
                        help='Path to the data configuration file')

    parser.add_argument('-s',
                        '--step',
                        type=str,
                        nargs='*',
                        required=True,
                        dest='pipeline_step',
                        help='Pipeline step to perform',
                        default=['preprocess', 'train', 'predict'],
                        choices=['preprocess', 'train', 'predict'])

    parser.add_argument('-m',
                        '--model-filename',
                        type=str,
                        required=False,
                        default=None,
                        dest='model_filename',
                        help='Path to model file')

    parser.add_argument('-o',
                        '--output-dir',
                        type=str,
                        default=None,
                        required=False,
                        dest='output_dir',
                        help='Path to output directory')

    parser.add_argument('-r',
                        '--regex-list',
                        type=str,
                        nargs='*',
                        required=False,
                        dest='inference_regex_list',
                        help='Inference regex list',
                        default=['*.tif'])

    parser.add_argument('-ib',
                        '--input-bands',
                        type=str,
                        nargs='*',
                        required=False,
                        dest='input_bands',
                        help='Inputs bands from incoming data',
                        default=None)

    parser.add_argument('-ob',
                        '--output-bands',
                        type=str,
                        nargs='*',
                        required=False,
                        dest='output_bands',
                        help='Output bands from incoming data',
                        default=None)

    parser.add_argument('-ps',
                        '--postprocessing-steps',
                        type=str,
                        nargs='*',
                        required=False,
                        dest='postprocessing_steps',
                        help='Postprocessing steps to perform',
                        default=['sieve', 'smooth', 'fill', 'dilate'],
                        choices=['sieve', 'smooth', 'fill', 'dilate'])

    args = parser.parse_args()

    # Setup timer to monitor script execution time
    timer = time.time()

    # Initialize pipeline object
    pipeline = CloudMaskPipeline(
        args.config_file,
        args.data_csv,
        args.model_filename,
        args.output_dir,
        args.inference_regex_list,
        args.input_bands,
        args.output_bands,
        args.postprocessing_steps
    )

    # Regression CHM pipeline steps
    if "preprocess" in args.pipeline_step:
        pipeline.preprocess()
    if "train" in args.pipeline_step:
        pipeline.train()
    if "predict" in args.pipeline_step:
        pipeline.predict()

    logging.info(f'Took {(time.time()-timer)/60.0:.2f} min.')

    return


# -----------------------------------------------------------------------------
# Invoke the main
# -----------------------------------------------------------------------------
if __name__ == "__main__":
    sys.exit(main())
